.wy-nav-content {
    max-width: 100% !important;
    padding: 15px !important;
}

.rst-content dl:not(.docutils) {
    margin: 0px 0px 0px 0px !important;
}

.data dd {
    margin-bottom: 0px !important;
}

.data .descname {
    border-right:10px !important;
}

.local-toc li ul li {
    padding-left: 20px !important;
}

.rst-content ul p {
    margin-bottom: 0px !important;
}

.rst-content li {
    list-style: outside;
    margin-left: 15px;
}

.document td {
    padding-bottom: 0px !important;
}

.document th {
    padding-top: 0px !important;
    padding-bottom: 0px !important;
}

.document th p {
    margin-bottom: 0px !important;
}

.document th p {
}

.function .descclassname {
    font-weight: normal !important;
}

.class .descclassname {
    font-weight: normal !important;
}

.admonition.warning {
    padding-top: 2px !important;
    padding-bottom: 2px !important;
}

.admonition.note {
    padding-top: 2px !important;
    padding-bottom: 2px !important;
}

.rst-content dl:not(.docutils) dt {
    color: #555;
}

.sig-paren {
    padding-left: 2px;
    padding-right: 2px;
}

h1, h2, h3 {
    background: #eee;
    padding: 5px;
    border-bottom: 1px solid #ccc;
}

h1 {
    font-size: 35px;
}

.admonition.warning {
    padding-top: 5px !important;
    padding-bottom: 5px !important;
}

.admonition.warning p {
    margin-bottom: 5px !important;
}

.admonition.note {
    padding-top: 5px !important;
    padding-bottom: 5px !important;
}

.admonition.note p {
    margin-bottom: 5px !important;
    backround-color: rgb(238, 255, 204) !important;
}

.codeblock div[class^='highlight'], pre.literal-block div[class^='highlight'], .rst-content .literal-block div[class^='highlight'], div[class^='highlight'] div[class^='highlight'] {
    background-color: #eeffcc !important;
}

.highlight .hll {
    background-color: #ffffcc
}

.highlight {
    background: #eeffcc;
}

.highlight-default, .highlight-python {
    border-radius: 3px !important;
    border: 1px solid #ac9 !important;
}

.highlight .c {
    color: #408090;
    font-style: italic
}

.wy-side-nav-search {
    background-color: grey !important
}

.highlight {
  border-radius: 3px !important;

}

div.highlight-default {
    margin-bottom: 10px !important;
}

pre {
    padding: 5px !important;
}

/* ================================================================== */
/* Warnings and info boxes like python doc */
/* ================================================================== */

div.admonition {
    margin-top: 10px !important;
    margin-bottom: 10px !important;
}

div.warning {
    background-color: #ffe4e4 !important;
    border: 1px solid #f66 !important;
    border-radius: 3px !important;
}

div.note {
    background-color: #eee !important;
    border: 1px solid #ccc !important;
    border-radius: 3px !important;
}

div.admonition p.admonition-title + p {
    display: inline !important;
}

p.admonition-title {
    display: inline !important;
    background: none !important;
    color: black !important;
}

p.admonition-title:after {
    content: ":" !important;
}

div.body div.admonition, div.body div.impl-detail {
}

.fa-exclamation-circle:before, .wy-inline-validate.wy-inline-validate-warning .wy-input-context:before, .wy-inline-validate.wy-inline-validate-info .wy-input-context:before, .rst-content .admonition-title:before {
    display: none !important;
}

.note code {
    background: #d6d6d6 !important;
}

/* ================================================================== */
/* Syntax highlight like Python doc.
/* ================================================================== */

/* Comment */
.highlight .err {
    border: 1px solid #FF0000
}

/* Error */
.highlight .k {
    color: #007020;
    font-weight: bold
}

/* Keyword */
.highlight .o {
    color: #666666
}

/* Operator */
.highlight .ch {
    color: #408090;
    font-style: italic
}

/* Comment.Hashbang */
.highlight .cm {
    color: #408090;
    font-style: italic
}

/* Comment.Multiline */
.highlight .cp {
    color: #007020
}

/* Comment.Preproc */
.highlight .cpf {
    color: #408090;
    font-style: italic
}

/* Comment.PreprocFile */
.highlight .c1 {
    color: #408090;
    font-style: italic
}

/* Comment.Single */
.highlight .cs {
    color: #408090;
    background-color: #fff0f0
}

/* Comment.Special */
.highlight .gd {
    color: #A00000
}

/* Generic.Deleted */
.highlight .ge {
    font-style: italic
}

/* Generic.Emph */
.highlight .gr {
    color: #FF0000
}

/* Generic.Error */
.highlight .gh {
    color: #000080;
    font-weight: bold
}

/* Generic.Heading */
.highlight .gi {
    color: #00A000
}

/* Generic.Inserted */
.highlight .go {
    color: #333333
}

/* Generic.Output */
.highlight .gp {
    color: #c65d09;
    font-weight: bold
}

/* Generic.Prompt */
.highlight .gs {
    font-weight: bold
}

/* Generic.Strong */
.highlight .gu {
    color: #800080;
    font-weight: bold
}

/* Generic.Subheading */
.highlight .gt {
    color: #0044DD
}

/* Generic.Traceback */
.highlight .kc {
    color: #007020;
    font-weight: bold
}

/* Keyword.Constant */
.highlight .kd {
    color: #007020;
    font-weight: bold
}

/* Keyword.Declaration */
.highlight .kn {
    color: #007020;
    font-weight: bold
}

/* Keyword.Namespace */
.highlight .kp {
    color: #007020
}

/* Keyword.Pseudo */
.highlight .kr {
    color: #007020;
    font-weight: bold
}

/* Keyword.Reserved */
.highlight .kt {
    color: #902000
}

/* Keyword.Type */
.highlight .m {
    color: #208050
}

/* Literal.Number */
.highlight .s {
    color: #4070a0
}

/* Literal.String */
.highlight .na {
    color: #4070a0
}

/* Name.Attribute */
.highlight .nb {
    color: #007020
}

/* Name.Builtin */
.highlight .nc {
    color: #0e84b5;
    font-weight: bold
}

/* Name.Class */
.highlight .no {
    color: #60add5
}

/* Name.Constant */
.highlight .nd {
    color: #555555;
    font-weight: bold
}

/* Name.Decorator */
.highlight .ni {
    color: #d55537;
    font-weight: bold
}

/* Name.Entity */
.highlight .ne {
    color: #007020
}

/* Name.Exception */
.highlight .nf {
    color: #06287e
}

/* Name.Function */
.highlight .nl {
    color: #002070;
    font-weight: bold
}

/* Name.Label */
.highlight .nn {
    color: #0e84b5;
    font-weight: bold
}

/* Name.Namespace */
.highlight .nt {
    color: #062873;
    font-weight: bold
}

/* Name.Tag */
.highlight .nv {
    color: #bb60d5
}

/* Name.Variable */
.highlight .ow {
    color: #007020;
    font-weight: bold
}

/* Operator.Word */
.highlight .w {
    color: #bbbbbb
}

/* Text.Whitespace */
.highlight .mb {
    color: #208050
}

/* Literal.Number.Bin */
.highlight .mf {
    color: #208050
}

/* Literal.Number.Float */
.highlight .mh {
    color: #208050
}

/* Literal.Number.Hex */
.highlight .mi {
    color: #208050
}

/* Literal.Number.Integer */
.highlight .mo {
    color: #208050
}

/* Literal.Number.Oct */
.highlight .sa {
    color: #4070a0
}

/* Literal.String.Affix */
.highlight .sb {
    color: #4070a0
}

/* Literal.String.Backtick */
.highlight .sc {
    color: #4070a0
}

/* Literal.String.Char */
.highlight .dl {
    color: #4070a0
}

/* Literal.String.Delimiter */
.highlight .sd {
    color: #4070a0;
    font-style: italic
}

/* Literal.String.Doc */
.highlight .s2 {
    color: #4070a0
}

/* Literal.String.Double */
.highlight .se {
    color: #4070a0;
    font-weight: bold
}

/* Literal.String.Escape */
.highlight .sh {
    color: #4070a0
}

/* Literal.String.Heredoc */
.highlight .si {
    color: #70a0d0;
    font-style: italic
}

/* Literal.String.Interpol */
.highlight .sx {
    color: #c65d09
}

/* Literal.String.Other */
.highlight .sr {
    color: #235388
}

/* Literal.String.Regex */
.highlight .s1 {
    color: #4070a0
}

/* Literal.String.Single */
.highlight .ss {
    color: #517918
}

/* Literal.String.Symbol */
.highlight .bp {
    color: #007020
}

/* Name.Builtin.Pseudo */
.highlight .fm {
    color: #06287e
}

/* Name.Function.Magic */
.highlight .vc {
    color: #bb60d5
}

/* Name.Variable.Class */
.highlight .vg {
    color: #bb60d5
}

/* Name.Variable.Global */
.highlight .vi {
    color: #bb60d5
}

/* Name.Variable.Instance */
.highlight .vm {
    color: #bb60d5
}

/* Name.Variable.Magic */
.highlight .il {
    color: #208050
}

.rst-content pre.literal-block, .rst-content div[class^='highlight'] {
    border: 1px solid #e1e4e5;
    padding: 0px;
    overflow-x: auto;
    margin: 1px 0 0px 0;
}
